#include <linux/fs.h>              // file system
#include <linux/module.h>          // module config
#include <linux/kernel.h>          // kernel
#include <linux/init.h>            // init exit 
#include <linux/moduleparam.h>     // 
#include <linux/list.h>            // 
#include <linux/cdev.h>            // cdev_register
#include <linux/mm.h>              // 
#include <linux/seq_file.h>        // 
#include <linux/ioport.h>          // 
#include <linux/delay.h>           // for delay
#include <asm/io.h>                // ioremap 
#include <linux/io.h>              // 
#include <linux/device.h>          // dev_t
#include <linux/platform_device.h> // 
#include <linux/syscalls.h>        // 
#include <asm/gpio.h>              // gpio

#include <linux/completion.h>
#include <linux/dmaengine.h>
#include <linux/dma-mapping.h>
#include <linux/platform_data/edma.h>
#include <linux/version.h>

// #include "edmatest.h"
#include "drvr.h"
#include "generic.h"

#define FPGA_NRST_H     gpio_set_value(90, 1);
#define FPGA_NRST_L     gpio_set_value(90, 0);

#define USER_LED1_H     gpio_set_value(115, 1);
#define USER_LED1_L     gpio_set_value(115, 0);

//auto detect Linux DMA Engine API (Kernel 4.4+)
#ifndef EDMA_DMA_COMPLETE
#define USE_DMA_ENGINE
#endif

static volatile int irqraised1;
static dma_addr_t dmaphysbuf;
static struct completion dma_comp;
#ifdef USE_DMA_ENGINE
static dma_cookie_t cookie;
#endif

struct drvr_mem et_mem;

/* 设备驱动相关的结构体 */
struct dev_struct
{    
    struct cdev cdev;                   /* 字符设备 */
    dev_t  devno;                       /* 设备号   */
    struct class *class;                /* class 类 */
    struct fasync_struct *async_queue;  /* 异步通知 */
    char   *user_buff;                  /* 用户缓存 */
    struct semaphore sem;               /* 原子操作 */
};
static struct dev_struct et_struct;

#ifdef USE_DMA_ENGINE
static void dma_callback(void *param)
{
    struct drvr_mem *mem_dev = (struct drvr_mem*) param;
    struct dma_chan *chan = mem_dev->dma.chan;

    switch (dma_async_is_tx_complete(chan, cookie, NULL, NULL)) {
        case DMA_COMPLETE:
            irqraised1 = 1;
            DBG_LOG("edma completed !\n");
            break;

        case DMA_ERROR:
            irqraised1 = -1;
            break;

        default:
            irqraised1 = -1;
            break;
    }

    complete(&dma_comp);
}
#else
// static void dma_callback(unsigned lch, u16 ch_status, void *data)
// {
    // switch (ch_status) {
        // case EDMA_DMA_COMPLETE:
            // irqraised1 = 1;
            // break;

        // case EDMA_DMA_CC_ERROR:
            // irqraised1 = -1;
            // break;

        // default:
            // irqraised1 = -1;
            // break;
    // }

    // complete(&dma_comp);
// }
#endif /* USE_DMA_ENGINE */

void et_dma_init(void)
{
    /*  completion是一种轻量级的机制，它允许一个线程告诉另一个线程工作已经完成。 */
    init_completion(&dma_comp);
}

int et_dma_open(struct drvr_mem* mem_dev, dma_addr_t *physbuf)
{
#ifdef USE_DMA_ENGINE
    struct dma_slave_config    conf;
    dma_cap_mask_t mask;
#endif

    /* Allocate DMA buffer */
    mem_dev->dma.buf = dma_alloc_coherent(NULL, MAX_DMA_TRANSFER_IN_BYTES,
                          &dmaphysbuf, 0);

    if (!mem_dev->dma.buf) {
        DBG_LOG("failed to allocate DMA buffer\n");

        return -ENOMEM;
    }

    *physbuf = dmaphysbuf;

#ifdef USE_DMA_ENGINE
    /* Allocate DMA channel */
    dma_cap_zero(mask);
    dma_cap_set(DMA_MEMCPY, mask);
    mem_dev->dma.chan = dma_request_channel(mask, NULL, NULL);

    if (!mem_dev->dma.chan) {
        return -ENODEV;
    }

    /* Configure DMA channel */
    conf.direction = DMA_MEM_TO_MEM;
    /* conf.dst_addr_width = DMA_SLAVE_BUSWIDTH_2_BYTES; */
    dmaengine_slave_config(mem_dev->dma.chan, &conf);

    DBG_LOG("Using Linux DMA Engine API");
#else
    // mem_dev->dma.dma_chan = edma_alloc_channel(EDMA_CHANNEL_ANY, dma_callback, NULL, EVENTQ_0);

    // if (mem_dev->dma.dma_chan < 0) {
        // DBG_LOG("edma_alloc_channel failed for dma_ch, error: %d\n", mem_dev->dma.dma_chan);

        // return mem_dev->dma.dma_chan;
    // }

    // DBG_LOG("Using EDMA/DMA Engine");
#endif /* USE_DMA_ENGINE */

    DBG_LOG("EDMA channel %d reserved\n", mem_dev->dma.dma_chan);

    return 0;
}

void et_dma_release(struct drvr_mem* mem_dev)
{
#ifdef USE_DMA_ENGINE
    dma_release_channel(mem_dev->dma.chan);
#else
    // edma_free_channel(mem_dev->dma.dma_chan);
#endif /* USE_DMA_ENGINE */
    dma_free_coherent(NULL, MAX_DMA_TRANSFER_IN_BYTES, mem_dev->dma.buf, dmaphysbuf);
}

int et_dma_copy(struct drvr_mem* mem_dev, unsigned long trgt_addr, unsigned long src_addr, int count)
{
    int result = 0;

#ifdef USE_DMA_ENGINE
    struct dma_chan *chan;
    struct dma_device *dev;
    struct dma_async_tx_descriptor *tx;
    unsigned long flags;

    chan = mem_dev->dma.chan;
    dev = chan->device;
    flags = DMA_CTRL_ACK | DMA_PREP_INTERRUPT;
    tx = dev->device_prep_dma_memcpy(chan, trgt_addr, src_addr, count, flags);

    if (!tx) {
        DBG_LOG("device_prep_dma_memcpy failed\n");

        return -ENODEV;
    }

    irqraised1 = 0u;
    dma_comp.done = 0;
    /* set the callback and submit the transaction */
    tx->callback = dma_callback;
    tx->callback_param = mem_dev;
    cookie = dmaengine_submit(tx);
    dma_async_issue_pending(chan);
#else
    // struct edmacc_param param_set;
    // int dma_ch = mem_dev->dma.dma_chan;

    // edma_set_src(dma_ch, src_addr, INCR, W256BIT);
    // edma_set_dest(dma_ch, trgt_addr, INCR, W256BIT);
    // edma_set_src_index(dma_ch, 1, 1);
    // edma_set_dest_index(dma_ch, 1, 1);
    // /* A Sync Transfer Mode */
    // edma_set_transfer_params(dma_ch, count, 1, 1, 1, ASYNC);//one block of one frame of one array of count bytes

    // /* Enable the Interrupts on Channel 1 */
    // edma_read_slot(dma_ch, &param_set);
    // param_set.opt |= ITCINTEN;
    // param_set.opt |= TCINTEN;
    // param_set.opt |= EDMA_TCC(EDMA_CHAN_SLOT(dma_ch));
    // edma_write_slot(dma_ch, &param_set);
    // irqraised1 = 0u;
    // dma_comp.done = 0;
    // result = edma_start(dma_ch);

    // if (result != 0) {
        // DBG_LOG("edma copy failed\n");

        // return result;
    // }
#endif /* USE_DMA_ENGINE */

    wait_for_completion(&dma_comp);

    /* Check the status of the completed transfer */

    if (irqraised1 < 0) {
        DBG_LOG("edma copy: Event Miss Occured!!!\n");
#ifdef USE_DMA_ENGINE
        dmaengine_terminate_all(chan);
#else
        // edma_stop(dma_ch);
#endif /* USE_DMA_ENGINE */
        result = -EAGAIN;
    }

    return result;
}

// static struct file_operations fpga_fops = 
// {
    // .owner   = THIS_MODULE,
    // .open    = fpga_open,
    // .mmap    = fpga_mmap,
    // .release = fpga_close,
    // .write   = fpga_write,
    // .read    = fpga_read,
    // .fasync  = fpga_fasync,
    // .unlocked_ioctl = fpga_ioctl,
// };

static  int __init et_init(void)
{
    int  rtn;
    unsigned long src_addr, dst_addr;
#if 0
    /* 自动分配主设备号 */
    rtn = alloc_chrdev_region( &(et_struct.devno), 0, 1, DEVICENAME);
    if (rtn < 0) {
        DBG_LOG("alloc_chrdev_region() failed: %d\n", rtn);
        return rtn;
    }
    
    // DBG_LOG( "fpga_device major#: %d, minor#: %d\n", 
    //         MAJOR(et_struct.devno), 
    //         MINOR(et_struct.devno) );

    /* 初始化 cdev 成员，建立 file_ops 连接 */
    cdev_init(&(et_struct.cdev), &fpga_fops);
    et_struct.cdev.owner = THIS_MODULE;
    /* 注册设备，是系统知道模块存在 */
    rtn = cdev_add( &(et_struct.cdev), et_struct.devno, 1);
    if (rtn) {
        DBG_LOG("cdev_add() failed: %d\n", rtn);
        unregister_chrdev_region( et_struct.devno, 1);
        return rtn;
    }
    
    et_struct.class = class_create(THIS_MODULE, DEVICENAME);                   /* 创建类 */
    device_create(et_struct.class, NULL, et_struct.devno, NULL, DEVICENAME);   /* 创建节点 */
#endif
    
    /* edmatest 初始化 */
    et_dma_init();
    
    /* 映射目标地内存 */
    // if (request_mem_region((unsigned int)FPGA_ADDR, FPGA_SIZE, DEVICENAME) == NULL) {
        // DBG_LOG("Failed to request fpga I/O memory region\n");
        // return -ENOMEM;
    // }
    // virtaddr_dst = ioremap(((unsigned int)FPGA_ADDR), FPGA_SIZE);
    // if (virtaddr_dst == NULL) {
        // DBG_LOG("Failed to fpga remap I/O memory\n");
        // return -ENOMEM;
    // }
    
    /* 映射源地址内存 */
    if (request_mem_region((unsigned int)GPMC_ADDR, GPMC_SIZE, DEVICENAME) == NULL) {
        DBG_LOG("failed to request gpmc I/O memory region\n");
        return -ENOMEM;
    }
    et_mem.virt_addr = ioremap(((unsigned int)GPMC_ADDR), GPMC_SIZE);
    if (et_mem.virt_addr == NULL) {
        DBG_LOG("failed to remap gpmc I/O memory\n");
        return -ENOMEM;
    }
    rtn = et_dma_open(&et_mem, &dmaphysbuf);
    
    /* 初始化成功 */
    DBG_LOG("init !\n");
    
    /* 开始传输 */
    src_addr = (unsigned long)GPMC_ADDR;
    dst_addr = (unsigned long)dmaphysbuf;
    
    rtn = et_dma_copy(&et_mem, dst_addr, src_addr, MAX_DMA_TRANSFER_IN_BYTES);
    if (rtn < 0) {
        DBG_LOG("failed to trigger EDMA transfer\n");
        return rtn;
    }
    
    /* 校验数据 */
    
    return 0;
}

static void __exit et_exit(void)
{
    release_mem_region((unsigned int)GPMC_ADDR, GPMC_SIZE);
    iounmap(et_mem.virt_addr);
    
#if 0
    if (et_struct.user_buff) {
        kfree(et_struct.user_buff);
    }
    
    device_destroy(et_struct.class, et_struct.devno);    /* 删除节点 */
    class_destroy(et_struct.class);                      /* 删除类 */
    cdev_del(&(et_struct.cdev));
    unregister_chrdev_region(et_struct.devno, 1);

    /* 注销设备 */
    unregister_chrdev_region(et_struct.devno, 1);
#endif
    
    DBG_LOG("exit !\n");
}

module_init(et_init);
module_exit(et_exit);
MODULE_LICENSE ("GPL");
